<?php

require_once("binary.php");

/*****************************************************************************/
function decode_2bpp_tiles ($f, &$p, $n_tiles)
/*****************************************************************************/
{
	/* Description: Decode 2-bpp (4-color) tiles.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $n_tiles: Number of tiles.
	 * Returns: The decoded tiles in SNES planar format.
	 */

	$tiles = "";
	$end = false;

	for ($tile_index = 0; $tile_index < $n_tiles; $tile_index++)
	{
		$tile = str_repeat ("\x00", 16);
		$method = getint1 ($f, $p);
		switch ($method)
		{
			case 0x00:
				$tile = getstr($f, $p, 16); // raw copy
				break;
			case 0x01:
				fill_2bpp_same ($tile, $f, $p);
				break;
			case 0x02:
				fill_2bpp_diff ($tile, $f, $p);
				break;
			case 0x03:
				fill_masked_rem_2bpp_same ($tile, $f, $p, 0); // read_2bpp_same
				break;
			case 0x04:
				fill_masked_rem_2bpp_same ($tile, $f, $p, 1);
				break;
			case 0x05:
				fill_masked_rem_2bpp_same ($tile, $f, $p, 2);
				break;
			case 0x06:
				fill_masked_rem_2bpp_diff ($tile, $f, $p, 1);
				break;
			case 0x07:
				fill_masked_rem_2bpp_diff ($tile, $f, $p, 2);
				break;
			case 0x08:
				fill_masked_rem_2bpp_diff ($tile, $f, $p, 3);
				break;
			case 0x09:
				fill_masked_rem_2bpp_diff ($tile, $f, $p, 4);
				break;
			case 0x0A:
				fill_masked_rem_plane ($tile, $f, $p, 0, 0);
				fill_masked_rem_plane ($tile, $f, $p, 1, 2);
				break;
			case 0x0B:
				fill_masked_rem_plane ($tile, $f, $p, 0, 1);
				fill_masked_rem_plane ($tile, $f, $p, 1, 2);
				break;
			case 0x0C:
				fill_masked_rem_plane ($tile, $f, $p, 0, 2);
				fill_masked_rem_plane ($tile, $f, $p, 1, 0);
				break;
			case 0x0D:
				fill_masked_rem_plane ($tile, $f, $p, 0, 2);
				fill_masked_rem_plane ($tile, $f, $p, 1, 1);
				break;
			case 0x0E:
				fill_masked_rem_plane ($tile, $f, $p, 0, 2);
				fill_masked_rem_plane ($tile, $f, $p, 1, 2);
				break;
			//case 0x0F:
			// (not used)
			//	break;
			case 0x10:
				fill_masked_flat ($tile, $f, $p, "\x00");
				break;
			case 0x11:
				fill_masked_flat ($tile, $f, $p, "\x55");
				break;
			case 0x12:
				fill_masked_flat ($tile, $f, $p, "\xAA");
				break;
			case 0x13:
				fill_masked_flat ($tile, $f, $p, "\xFF");
				break;
			case 0x14:
				copy_tile_masked_2bpp ($tile, $f, $p, $tiles);
				break;
			default:
				print "Unknown 2-bpp tile compression: " . dechex ($method) . " at offset 0x" . dechex ($p) . "\n";
				$end = true;
		}
		if ($end) break;
		$tiles .= $tile;
	}

	return ($tiles);
}

/*****************************************************************************/
function decode_4bpp_tiles ($f, &$p, $n_tiles)
/*****************************************************************************/
{
	/* Description: Decode 4-bpp (16-color) tiles.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $n_tiles: Number of tiles.
	 * Returns: The decoded tiles in SNES planar format.
	 */

	$tiles = "";
	$end = false;

	for ($tile_index = 0; $tile_index < $n_tiles; $tile_index++)
	{
		$tile = str_repeat ("\x00", 32);
		$method = getint1 ($f, $p);
		switch ($method)
		{
			case 0x00:
				$tile = getstr($f, $p, 32); // raw copy
				break;
			case 0x01:
				fill_4bpp_same ($tile, $f, $p);
				break;
			case 0x02:
				fill_4bpp_diff ($tile, $f, $p);
				break;
			case 0x03:
				fill_masked_rem_4bpp_diff ($tile, $f, $p, 1);
				break;
			case 0x04:
				fill_masked_rem_4bpp_diff ($tile, $f, $p, 2);
				break;
			case 0x05:
				fill_masked_rem_4bpp_diff ($tile, $f, $p, 3);
				break;
			case 0x06:
				fill_masked_rem_4bpp_diff ($tile, $f, $p, 4);
				break;
			case 0x07:
				fill_with_color ($tile, $f, $p);
				break;
			case 0x08:
				read_vsp ($tile, $f, $p);
				break;
			case 0x09:
			case 0x0A:
			case 0x0B:
			case 0x0C:
			case 0x0D:
			case 0x0E:
			case 0x0F:
			case 0x10:
			case 0x11:
			case 0x12:
			case 0x13:
			case 0x14:
			case 0x15:
			case 0x16:
			case 0x17:
				read_planes_masked ($tile, $f, $p, $method - 8);
				break;
			case 0x18:
				$copied_tile_index = getint1 ($f, $p);
				copy_tile_masked_4bpp ($tile, $f, $p, $tiles, $copied_tile_index);
				break;
			case 0x19:
				$copied_tile_index = getint2 ($f, $p);
				copy_tile_masked_4bpp ($tile, $f, $p, $tiles, $copied_tile_index);
				break;
			default:
				print "Unknown 4-bpp tile compression: " . dechex ($method) . " at offset 0x" . dechex ($p) . "\n";
				$end = true;
		}
		if ($end) break;
		$tiles .= $tile;
	}

	return ($tiles);
}

/*****************************************/
/*** Helper functions for 2-bpp tiles. ***/
/*****************************************/

/*****************************************************************************/
function fill_masked_flat (&$tile, $f, &$p, $fill_value)
/*****************************************************************************/
{
	/* Description: ???
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $fill_value: ???
	 * Returns: Nothing.
	 */

	$mask = getint2 ($f, $p);

	for ($i = 0; $i < 16; $i++)
	{
		if ($mask & (1 << $i))
		{
			$tile[$i] = $fill_value;
		} else {
			$tile[$i] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_masked_rem_2bpp_diff (&$tile, $f, &$p, $fill_count)
/*****************************************************************************/
{
	/* Description: Fill a part of the tile by repeating two bytes (one for each plane), then read the rest.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $fill_count: ???
	 * Returns: Nothing.
	 */

	$remain_mask = 0;
	// Filling steps.
	for ($fill_index = 0; $fill_index < $fill_count; $fill_index++) {
		$mask = getint1 ($f, $p);
		$x = getch ($f, $p);
		$y = getch ($f, $p);

		for ($i = 0; $i < 8; $i++)
		{
			if ($mask & (1 << $i))
			{
				$tile[$i * 2    ] = $x;
				$tile[$i * 2 + 1] = $y;
			}
		}
		$remain_mask |= $mask;
	}

	$mask = ~$remain_mask;
	// Read remainder.
	for ($i = 0; $i < 8; $i++)
	{
		if ($mask & (1 << $i))
		{
			$tile[$i * 2    ] = getch ($f, $p);
			$tile[$i * 2 + 1] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_masked_rem_2bpp_same (&$tile, $f, &$p, $fill_count)
/*****************************************************************************/
{
	/* Description: Fill a part of the tile by repeating a byte, then read the rest.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $fill_count: ???
	 * Returns: Nothing.
	 */

	$remain_mask = 0;
	// Filling steps.
	for ($fill_index = 0; $fill_index < $fill_count; $fill_index++) {
		$mask = getint1 ($f, $p);
		$fill_value = getch ($f, $p);

		for ($i = 0; $i < 8; $i++)
		{
			if ($mask & (1 << $i))
			{
				$tile[$i * 2] = $tile[$i * 2 + 1] = $fill_value;
			}
		}
		$remain_mask |= $mask;
	}

	$mask = ~$remain_mask;
	// Read remainder.
	for ($i = 0; $i < 8; $i++)
	{
		if ($mask & (1 << $i))
		{
			$tile[$i * 2] = $tile[$i * 2 + 1] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_masked_rem_plane (&$tile, $f, &$p, $plane_offset, $fill_count)
/*****************************************************************************/
{
	/* Description: ???
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $plane_offset: Offset of the plane.
	 * Returns: Nothing.
	 */

	$remain_mask = 0;
	// Filling steps.
	for ($fill_index = 0; $fill_index < $fill_count; $fill_index++) {
		$mask = getint1 ($f, $p);
		$fill_value = getch ($f, $p);

		for ($i = 0; $i < 8; $i++)
		{
			if ($mask & (1 << $i))
			{
				$tile[$i * 2 + $plane_offset] = $fill_value;
			}
		}
		$remain_mask |= $mask;
	}

	$mask = ~$remain_mask;
	// Read remainder.
	for ($i = 0; $i < 8; $i++)
	{
		if ($mask & (1 << $i))
		{
			$tile[$i * 2 + $plane_offset] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function copy_tile_masked_2bpp (&$tile, $f, &$p, $tiles)
/*****************************************************************************/
{
	/* Description: Copy an earlier tile and add some differences.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $tiles: Earlier tiles.
	 * Returns: Nothing.
	 */

	// Copy earlier tile.
	$tile_index = getint2 ($f, $p);
	$tile = substr ($tiles, $tile_index * 16, 16);
	// Apply differences.
	$mask = getint1 ($f, $p);
	for ($i = 0; $i < 8; $i++)
	{
		if (!($mask & (1 << $i)))
		{
			$tile[$i * 2    ] = getch ($f, $p);
			$tile[$i * 2 + 1] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_2bpp_diff (&$tile, $f, &$p)
/*****************************************************************************/
{
	/* Description: Fill the tile by repeating two bytes (one for each plane).
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * Returns: Nothing.
	 */

	$fill_value = getstr ($f, $p, 2);
	$tile = str_repeat($fill_value, 8);
}

/*****************************************************************************/
function fill_2bpp_same (&$tile, $f, &$p)
/*****************************************************************************/
{
	/* Description: Fill the tile by repeating a byte.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * Returns: Nothing.
	 */

	$fill_value = getch ($f, $p);
	$tile = str_repeat($fill_value, 16);
}

/*****************************************/
/*** Helper functions for 4-bpp tiles. ***/
/*****************************************/

/*****************************************************************************/
function copy_tile_masked_4bpp (&$tile, $f, &$p, $tiles, $tile_index)
/*****************************************************************************/
{
	/* Description: Copy an earlier tile and add some differences.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $tiles: Earlier tiles.
	 * - $tile_index: ???
	 * Returns: Nothing.
	 */

	// Copy earlier tile.
	$tile = substr ($tiles, $tile_index * 32, 32);
	// Apply differences.
	$mask = getint1 ($f, $p);
	for ($i = 0; $i < 8; $i++)
	{
		if (!($mask & (1 << $i)))
		{
			$tile[$i * 2     ] = getch ($f, $p);
			$tile[$i * 2 +  1] = getch ($f, $p);
			$tile[$i * 2 + 16] = getch ($f, $p);
			$tile[$i * 2 + 17] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_4bpp_same (&$tile, $f, &$p)
/*****************************************************************************/
{
	/* Description: Fill the tile by repeating a byte.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * Returns: Nothing.
	 */

	$fill_value = getch ($f, $p);
	$tile = str_repeat ($fill_value, 32);
}

/*****************************************************************************/
function read_plane_masked (&$tile, $f, &$p, $plane_offset)
/*****************************************************************************/
{
	/* Description: Fill a single plane of the tile, then read the other bytes.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $plane_offset: Offset of the plane.
	 * Returns: Nothing.
	 */

	$fill_value = getch ($f, $p);
	$mask = getint1 ($f, $p);

	for ($i = 0; $i < 8; $i++)
	{
		if ($mask & (1 << $i))
		{
			$tile[$i * 2 + $plane_offset] = $fill_value;
		} else {
			$tile[$i * 2 + $plane_offset] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function read_plane (&$tile, $f, &$p, $plane_offset)
/*****************************************************************************/
{
	/* Description: Read a single plane of the tile.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $plane_offset: Offset of the plane.
	 * Returns: Nothing.
	 */

	for ($i = 0; $i < 8; $i++)
	{
		$tile[$i * 2 + $plane_offset] = getch ($f, $p);
	}
}

/*****************************************************************************/
function read_planes_masked (&$tile, $f, &$p, $mask)
/*****************************************************************************/
{
	/* Description: ???
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $mask: ???
	 * Returns: Nothing.
	 */

	$planes = array(0, 1, 16, 17);
	foreach ($planes as $plane_index => $plane_offset)
	{
		if ($mask & (1 << $plane_index))
		{
			read_plane_masked ($tile, $f, $p, $plane_offset);
		} else {
			read_plane ($tile, $f, $p, $plane_offset);
		}
	}
}

/*****************************************************************************/
function fill_masked_rem_4bpp_diff (&$tile, $f, &$p, $fill_count)
/*****************************************************************************/
{
	/* Description: Fill a part of the tile by repeating four bytes (one for each plane), then read the rest.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $fill_count: ???
	 * Returns: Nothing.
	 */

	$remain_mask = 0;
	// Filling steps.
	for ($fill_index = 0; $fill_index < $fill_count; $fill_index++) {
		$mask = getint1 ($f, $p);
		$a = getch ($f, $p);
		$b = getch ($f, $p);
		$c = getch ($f, $p);
		$d = getch ($f, $p);

		for ($i = 0; $i < 8; $i++)
		{
			if ($mask & (1 << $i))
			{
				$tile[$i * 2     ] = $a;
				$tile[$i * 2 +  1] = $b;
				$tile[$i * 2 + 16] = $c;
				$tile[$i * 2 + 17] = $d;
			}
		}
		$remain_mask |= $mask;
	}

	$mask = ~$remain_mask;
	// Read remainder.
	for ($i = 0; $i < 8; $i++)
	{
		if ($mask & (1 << $i))
		{
			$tile[$i * 2     ] = getch ($f, $p);
			$tile[$i * 2 +  1] = getch ($f, $p);
			$tile[$i * 2 + 16] = getch ($f, $p);
			$tile[$i * 2 + 17] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_4bpp_diff (&$tile, $f, &$p)
/*****************************************************************************/
{
	/* Description: Fill the tile by repeating four bytes (one for each plane).
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * Returns: Nothing.
	 */

	$a = getch ($f, $p);
	$b = getch ($f, $p);
	$c = getch ($f, $p);
	$d = getch ($f, $p);

	for ($i = 0; $i < 8; $i++)
	{
		$tile[$i * 2     ] = $a;
		$tile[$i * 2 +  1] = $b;
		$tile[$i * 2 + 16] = $c;
		$tile[$i * 2 + 17] = $d;
	}
}

/*****************************************************************************/
function read_vsp (&$tile, $f, &$p)
/*****************************************************************************/
{
	/* Description: ???
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * Returns: Nothing.
	 */

	$mask = getint1 ($f, $p);

	for ($i = 0; $i < 8; $i++)
	{
		if ($mask & (1 << $i))
		{
			// Expand 4*2 bits to 4 bytes.
			// Multiplying a 2-bit number with 0x55 will repeat the 2 bits 4 times.
			$x = getint1 ($f, $p);
			$tile[$i * 2     ] = chr (($x      & 3) * 0x55);
			$tile[$i * 2 +  1] = chr (($x >> 2 & 3) * 0x55);
			$tile[$i * 2 + 16] = chr (($x >> 4 & 3) * 0x55);
			$tile[$i * 2 + 17] = chr (($x >> 6 & 3) * 0x55);
		} else {
			$tile[$i * 2     ] = getch ($f, $p);
			$tile[$i * 2 +  1] = getch ($f, $p);
			$tile[$i * 2 + 16] = getch ($f, $p);
			$tile[$i * 2 + 17] = getch ($f, $p);
		}
	}
}

/*****************************************************************************/
function fill_with_color (&$tile, $f, &$p)
/*****************************************************************************/
{
	/* Description: Fill the tile with a single color.
	 * Parameters:
	 * - $tile: The current tile.
	 * - $f: ROM image.
	 * - $p: Offset.
	 * Returns: Nothing.
	 */

	$color = getint1 ($f, $p);
	for ($i = 0; $i < 8; $i++)
	{
		$tile[$i * 2     ] = ($color & 1 ? "\xFF" : "\x00");
		$tile[$i * 2 +  1] = ($color & 2 ? "\xFF" : "\x00");
		$tile[$i * 2 + 16] = ($color & 4 ? "\xFF" : "\x00");
		$tile[$i * 2 + 17] = ($color & 8 ? "\xFF" : "\x00");
	}
}

?>