<?php

require_once("binary.php");

function planar_to_interleaved ($planar, $bpp, $plane_offsets, $tile_size, $row_stride)
{
	/* Description: Convert tiles from planar (SNES) to interleaved (BMP).
	 * Parameters:
	 * - $planar: The tiles in planar format, as the SNES uses them.
	 * - $bpp: Bits per pixel, i.e. the number of planes.
	 * - $plane_offsets: Offset of each plane from the beginning of the tile.
	 * - $tile_size: Offset of the beginning of each tile from the previous. (in bytes)
	 * - $row_stride: Offset of the beginning of each row from the previous. (in bytes)
	 * Returns: Tiles in 8-bpp interleaved format, as they need to be in a BMP file.
	 */

	$tile_width = $tile_height = 8; // tile width/height (in pixels)
	$in_pos = 0;
	$out_pos = 0;
	$n_tiles = (int)(strlen($planar) / $tile_size); // number of tiles
	$interleaved = str_repeat ("\x00", $n_tiles * $tile_width * $tile_height);

	for ($ix = 0; $ix < $n_tiles; $ix++) // for each tile
	{
		for ($y = 0; $y < $tile_height; $y++) // for each pixel row
		{
			for ($x = 0; $x < $tile_width; $x++) // for each pixel
			{
				$color = 0;
				for ($plane = 0; $plane < $bpp; $plane++) // for each bitplane
				{
					if (ord ($planar[$in_pos + ($y * $row_stride) + $plane_offsets[$plane]]) & (128 >> $x))
						{ $color |= 1 << $plane; }
				}
				$interleaved[$out_pos] = chr ($color);
				$out_pos++;
			}
		}
		$in_pos += $tile_size;
	}

	return $interleaved;
}

function make_snes_map ($map, $planar_tiles, $fn_bmp, $map_width, $map_height, $bpp, $pal, $clip_wh = -1, $clip_hh = -1)
{
	/* Description: Write a SNES map to a BMP file.
	 * Parameters:
	 * - $map: The tilemap in SNES format. (2 bytes per tilepos)
	 * - $planar_tiles: The tiles in SNES format.
	 * - $fn_bmp: Name of BMP file to create.
	 * - $map_width: Width of $map in tiles.
	 * - $map_height: Height of $map in tiles.
	 * - $bpp: Bits per pixel. (of $planar_tiles)
	 * - $pal: The palette in BMP format.
	 * - $clip_wh, $clip_hh: Clipping width and height. (optional, used only for sprites)
	 * Returns: Nothing.
	 */
//print "Convert tiles\n";
	// Convert tiles to BMP format.
	switch ($bpp) {
		case 1: // special format used by background sprite masks
			$plane_offsets = array(0x00);
			$row_stride = 1;
			$tile_size = 8;
		break;
		case 2: // standard 2-bpp SNES tiles (level layer 1, font)
			$plane_offsets = array(0x00, 0x01);
			$row_stride = 2;
			$tile_size = 16;
		break;
		case 4: // standard 4-bpp SNES tiles (level layer 2 and 3, screen, sprites)
			$plane_offsets = array(0x00, 0x01, 0x10, 0x11);
			$row_stride = 2;
			$tile_size = 32;
		break;
		case 5: // special format used by background sprites: 4 bits for color + 1 bit for masking
			$plane_offsets = array(1, 2, 3, 4);
			$bpp = 4; // first byte is some mask
			$row_stride = 5;
			$tile_size = 5*8;
		break;
		default:
			die("make_snes_map: Unknown bpp: " . $bpp);
		break;
	}
	$interleaved_tiles = planar_to_interleaved($planar_tiles, $bpp, $plane_offsets, $tile_size, $row_stride);
	$n_tiles = (int)(strlen($planar_tiles) / $tile_size); // number of tiles

	$tile_width = $tile_height = 8; // fixed

//print "Assemble image\n";
	// Assemble image from the tiles according to the tilemap.
	// This part takes the most time.
	$bm = str_repeat ("\x00", $map_width * $map_height * $tile_width * $tile_height);

	$image_width = $map_width * $tile_width;
	$image_height = $map_height * $tile_height;
/*
	$p_bm = 0;
	for ($y_tile = 0; $y_tile < $map_height; $y_tile++) // for each tile row
	{
		for ($y_pixel = 0; $y_pixel < $tile_height; $y_pixel++) // for each pixel row
		{
			$p_map = $y_tile * $map_width * 2;
			for ($x_tile = 0; $x_tile < $map_width; $x_tile++) // for each tile
			{
				$t = getint2 ($map, $p_map);
				$tile_index = $t & 0x3FF; // tile index
				$palette_base = (($t >> 10) & 7) << $bpp; // palette index
				if ($tile_index >= $n_tiles) { $tile_index = 0; } // tile index overflow
				$palette_base &= 0x3F;
				$x_flip = ($t & 0x4000) ? ($tile_width - 1) : 0; // flip horizontally
				$y_flip = ($t & 0x8000) ? ($tile_height - 1) : 0; // flip vertically
				//*** $prio = $t & 0x2000; *** / // in front of sprites

				for ($x_pixel = 0; $x_pixel < $tile_width; $x_pixel++) // for each pixel
				{
					$bm[$p_bm++] = $interleaved_tiles[$tile_index * $tile_width * $tile_height +
						($y_pixel ^ $y_flip) * $tile_width + ($x_pixel ^ $x_flip)] | chr ($palette_base);
				}
			}
		}
	}
*/
	// Is this faster?
	$ts = $tile_width * $tile_height; // tile size
	// jumps in the output:
	$e1 = $image_width - $tile_width; // end of pixel line in a tile
	$e2 = $tile_width - $tile_height * $image_width; // end of tile
	$e3 = $image_width * ($tile_height - 1); // end of a line of tiles

	$p_map = $p_bm = 0;
	for ($y_tile = 0; $y_tile < $map_height; $y_tile++) // for each tile row
	{
		for ($x_tile = 0; $x_tile < $map_width; $x_tile++) // for each tile
		{
			$t = getint2 ($map, $p_map);
			$tile_index = $t & 0x3FF; // tile index
			$palette_base = (($t >> 10) & 7) << $bpp; // palette index
			if ($tile_index >= $n_tiles) { $tile_index = 0; } // tile index overflow
			$palette_base &= 0x3F;
			$x_flip = ($t & 0x4000) ? ($tile_width - 1) : 0; // flip horizontally
			$y_flip = ($t & 0x8000) ? ($tile_height - 1) : 0; // flip vertically
			//*** $prio = $t & 0x2000; *** / // in front of sprites
			$flip = ($y_flip << 3) | $x_flip;

			$p_tile = $tile_index * $ts;
			for ($y_pixel = 0; $y_pixel < $tile_height; $y_pixel++) // for each pixel row
			{
				for ($x_pixel = 0; $x_pixel < $tile_width; $x_pixel++) // for each pixel
				{
					$bm[$p_bm] = $interleaved_tiles[$p_tile ^ $flip] | chr ($palette_base);
					$p_bm++;
					$p_tile++;
				}
				$p_bm += $e1;
			}
			$p_bm += $e2;
		}
		$p_bm += $e3;
	}

//print "Create BMP\n";
	// Create BMP.
	if ($clip_wh < 0) $clip_wh = $image_width;
	if ($clip_hh < 0) $clip_hh = $image_height;

	write_bmp ($fn_bmp, $clip_wh, $clip_hh, $bm, $pal, 8, $image_width);
//print "end\n";
}

?>