<?php

require_once("binary.php");
require_once("bmp.php");
require_once("compression_tiles.php");
require_once("mapping.php");
require_once("palettes.php");
require_once("compression_maps.php");
require_once("make_snes_map.php");
require_once("make_map.php");
require_once("compression_levels.php");
require_once("select_palette.php");

/*****************************************************************************/
function export_blocks ($f, &$p, $id, $bpp, $block_width, $block_height, $n_blocks, $n_tiles)
/*****************************************************************************/
{
	/* Description: Export level graphics (and the ending font).
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $id: Resource index.
	 * - $bpp: Bits per pixel of tiles. (2 or 4)
	 * - $block_width: Block width in tiles.
	 * - $block_height: Block height in tiles.
	 * - $n_blocks: Number of actual blocks.
	 * - $n_tiles: Number of actual tiles.
	 * Returns: Nothing.
	 */

	/**************/
	/*** BLOCKS ***/
	/**************/

	$n_stored_blocks = getint2 ($f, $p); // number of stored blocks
	// block mapping
	$blocks_mapping = decode_mapping($f, $p, 8, $n_blocks);

	// blocks
	$out_length = $n_stored_blocks * 2 * $block_width * $block_height;
	if ($block_width == 6) { $stored_blocks = decode_level_blocks_wide ($f, $p, $out_length); }
	elseif ($block_width == 3) { $stored_blocks = decode_level_blocks_narrow ($f, $p, $out_length); }
	else { die ('Illegal block_width ' . $block_width); }

	// apply block mapping
	$blocks = apply_mapping ($blocks_mapping, $stored_blocks, $block_width * $block_height * 2);

	/*************/
	/*** TILES ***/
	/*************/

	$n_stored_tiles = getint2 ($f, $p); // number of stored tiles
	if ($n_tiles == 0x100) { $value_bits = 8; }
	elseif ($n_tiles == 0x200) { $value_bits = 9; }
	else { die ('Illegal n_tiles ' . $n_tiles); }

	// tile mapping
	$tiles_mapping = decode_mapping($f, $p, $value_bits, $n_tiles);

	// tiles
	if ($bpp == 2) { $stored_tiles = decode_2bpp_tiles ($f, $p, $n_stored_tiles); }
	elseif ($bpp == 4) { $stored_tiles = decode_4bpp_tiles ($f, $p, $n_stored_tiles); }
	else { die ('Illegal bpp ' . $bpp); }

	// apply tile mapping
	$tiles = apply_mapping ($tiles_mapping, $stored_tiles, $bpp * 8);

	//TODO: There is something after this in layer 2.

	/**********************/
	/*** SELECT PALETTE ***/
	/**********************/

	if ($id == 74) {
		// ending font
		$pal_res = $pal_index = -1;
		$pal_length = 4;
	}
	else
	{
		// level graphics
		$pal_res = ($id + 3)&~3; // palette resource
		switch ($id & 3)
		{
			case 1: $pal_index = 0; $pal_length = 32; break;
			case 2: $pal_index = 6; $pal_length = 16; break;
			case 3: $pal_index = 2; $pal_length = 64; break;
		}
	}

	/********************/
	/*** FIND PALETTE ***/
	/********************/

	$pal = get_palette ($pal_res, $pal_index, $pal_length);

	/**********************/
	/*** CREATE TILEMAP ***/
	/**********************/

	// Arrange blocks in 16x16:
	$arrange_width = 16;
	$arrange_height = 16;
	$arrange_map = "";
	for ($i = 0; $i < 256; $i++)
	{
		$arrange_map .= chr ($i);
	}
	$map = make_map ($arrange_map, $blocks, $arrange_width, $arrange_height, $block_width * 2, $block_height, 0, 1, $iw, $ih);

	/******************/
	/*** CREATE BMP ***/
	/******************/

	$fn_bmp = sprintf ("exp0%s%02d.bmp", $GLOBALS['slash'], $id);
	make_snes_map ($map, $tiles, $fn_bmp, $iw/2, $ih, $bpp, $pal);
}

/*****************************************************************************/
function export_palette ($f, &$p, $id)
/*****************************************************************************/
{
	/* Description: Export a set of palettes as BMP.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $id: Resource index.
	 * Returns: Nothing.
	 */

	$height = 0;
	if ($id <= 40) $height = 8; // level graphics
	if ($id == 68) $height = 27; // sprite
	if ($id == 69) $height = 33; // screen
	$width = 16;

	$palette = read_palette ($f, $p, $height * $width);
	$fn_bmp = sprintf ("exp0%s%02d.bmp", $GLOBALS['slash'], $id);
	write_bmp ($fn_bmp, $width, $height, $palette, "", 32, $width * 4);
}

/*****************************************************************************/
function export_tiles ($f, &$p, $id)
/*****************************************************************************/
{
	/* Description: Export a set of tiles.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $id: Resource index.
	 * Returns: Nothing.
	 */

	/**********************/
	/*** SELECT PALETTE ***/
	/**********************/

	if ($id == 0) {
		// menu font
		$bpp = 2;
		$font_pal_p = 0x2130+7*4*2; // hardcoded palette in ROM (not in the resource table)
		$pal = read_palette ($f, $font_pal_p, 4);
		$length = 256;
		$width = 16;
		$height = 16;
	} else {
		$bpp = 4;
		$pal_index = -1; // palette index
		$pal_res = -1; // palette resource
		$length = 0;
		if ($id == 70) { $pal_index = 0; $pal_res = 69; $length = 4; } // hitpoint
		if ($id == 73) { $pal_index = 23; $pal_res = 68; $length = 16; } // princess room flame
		$pal = get_palette ($pal_res, $pal_index, 16);
		$width = 1;
		$height = $length;
	}

	$tiles = getstr ($f, $p, $length * 8 * $bpp);

	/**********************/
	/*** CREATE TILEMAP ***/
	/**********************/

	$map = "";
	for ($i = 0; $i < $length; $i++)
	{
		$map .= int22str($i);
	}

	/******************/
	/*** CREATE BMP ***/
	/******************/

	$fn_bmp = sprintf ("exp0%s%02d.bmp", $GLOBALS['slash'], $id);
	make_snes_map ($map, $tiles, $fn_bmp, $width, $height, $bpp, $pal);
}

/*****************************************************************************/
function export_screen ($f, &$p, $id)
/*****************************************************************************/
{
	/* Description: Export a full-screen image.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $id: Resource index.
	 * Returns: Nothing.
	 */

	/*******************/
	/*** SELECT SIZE ***/
	/*******************/

	if ($id >= 76 and $id <= 78 or $id == 84 or $id == 91)
	{
		$map_width = 64; $map_height = 56; // big maps (contains 2x2 screens)
	} else {
		$map_width = 32; $map_height = 28; // small maps (same size as the screen)
	}
	$map_size = $map_width * $map_height * 2; // size of map in bytes

	/******************/
	/*** DECOMPRESS ***/
	/******************/

	$f_map = decode_screen_tilemap ($f, $p, $map_size);
	$n_tiles = getint2 ($f, $p);
	$tiles = decode_4bpp_tiles ($f, $p, $n_tiles);

	/**********************/
	/*** SELECT PALETTE ***/
	/**********************/

	select_screen_palette ($id, $pal_res, $pal_index);

	/********************/
	/*** FIND PALETTE ***/
	/********************/

	$pal = get_palette ($pal_res, $pal_index, 64);

	/******************/
	/*** CREATE BMP ***/
	/******************/

	$fn_bmp = sprintf ("exp0%s%02d.bmp", $GLOBALS['slash'], $id);
	$bpp = 4;
	make_snes_map($f_map, $tiles, $fn_bmp, $map_width, $map_height, $bpp, $pal);
}

/*****************************************************************************/
function export_sprite ($f, &$p, $id)
/*****************************************************************************/
{
	/* Description: Export a sprite.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Start address.
	 * - $id: Index of resource.
	 * Returns: Nothing.
	 */

	/**********************/
	/*** SELECT PALETTE ***/
	/**********************/

	select_sprite_palette ($id, $pal_res, $pal_index);

	/********************/
	/*** FIND PALETTE ***/
	/********************/

	$pal = get_palette ($pal_res, $pal_index, 16);

	/*******************/
	/*** READ HEADER ***/
	/*******************/

	/*** read size ***/
	$rw = getint1 ($f, $p) + 1; // width in pixels
	$rh = getint1 ($f, $p) + 1; // height in pixels
	$x = getint1 ($f, $p);
	$hh = $x >> 4; // height in 16-pixel units
	$wh = $x & 0xF; // width in 16-pixel units

	/******************/
	/*** DECOMPRESS ***/
	/******************/

	$plane_offsets = array (0x00, 0x01, 0x10, 0x11, 0x20, 0x21, 0x30, 0x31); // for two 4-bpp tiles
	$tiles = "";
	for ($i = 0; $i < $wh * $hh * 2; $i++)
	{
		$mm = getint1 ($f, $p);
		$tile = str_repeat ("\x00", 64);
		for ($plane = 0; $plane < 8; $plane++)
		{
			if (!($mm & 1 << $plane)) { continue; }
			$plane_offset = $plane_offsets[$plane];
			$m = getint1 ($f, $p);
			if ($m == 255)
			{
				$m = getint1 ($f, $p);
				//$fill = true;
				$ch = getch ($f, $p);
				for ($b = 0; $b < 8; $b++)
				{
					if (!($m & 1 << $b)) { $tile[$b * 2 + $plane_offset] = $ch; }
				}
			} else {
				//$fill = false;
				for ($b = 0; $b < 8; $b++)
				{
					if (!($m & 1 << $b)) { $tile[$b * 2 + $plane_offset] = getch ($f, $p); }
				}
			}
		}
		$tiles .= $tile;
	}

	/**********************/
	/*** CREATE TILEMAP ***/
	/**********************/
	/*
	 * 0 1 | 4 5 | ... wrap when you reach the right edge
	 * 2 3 | 6 7 | ...
	 */
	$map = "";
	$k = 0;
	for ($y = 0; $y < $hh; $y++) // for each tile row
	{
		$row1 = $row2 = "";
		for ($x = 0; $x < $wh; $x++) // for each tile column
		{
			$row1 .= int22str($k);
			$row1 .= int22str($k + 1);
			$row2 .= int22str($k + 2);
			$row2 .= int22str($k + 3);
			$k += 4;
		}
		$map .= $row1;
		$map .= $row2;
	}

	/******************/
	/*** CREATE BMP ***/
	/******************/

	$fn_bmp = sprintf ('exp1%s%04d.bmp', $GLOBALS['slash'], $id);
	$bpp = 4;
	make_snes_map($map, $tiles, $fn_bmp, $wh * 2, $hh * 2, $bpp, $pal, $rw, $rh);
}

/*****************************************************************************/
function export_bg_sprite ($f, &$q, $id)
/*****************************************************************************/
{
	/* Description: Export a background sprite.
	 * Parameters:
	 * - $f: ROM image.
	 * - $q: Start address.
	 * - $id: Index of resource.
	 * Returns: Nothing.
	 */

	/**********************/
	/*** SELECT PALETTE ***/
	/**********************/

	$mono = false;

	if ($id >= 1395) { $mono = true; } /*** masks ***/
	else select_bg_sprite_palette ($id, $pal_res, $pal_index);

	/********************/
	/*** FIND PALETTE ***/
	/********************/

	if ($mono) {
		$pal = get_palette(-1, -1, 2);
	} else {
		$pal = get_palette($pal_res, $pal_index, 16);
	}

	/******************/
	/*** DECOMPRESS ***/
	/******************/

	$p = $q;
	$p0 = $p;

	// Read start addresses of columns. (6 columns, each is 8 pixels wide.)
	$wh = 6; // width in tiles
	$height_pixels = 0x40; // height in pixels
	$height_tiles = (int)($height_pixels / 8);
	$tiles = "";
	$bpp = ($mono ? 1 : 5);
	$empty_row = ($mono ? "\x00" : "\xFF\x00\x00\x00\x00");

	for ($i = 0; $i < $wh; $i++)
		{ $ad[$i] = getint2 ($f, $p); }

	// For each column:
	for ($i = 0; $i < $wh; $i++)
	{
		if (0 == $ad[$i])
		{
			// Empty column:
			$tile_column = str_repeat ($empty_row, $height_pixels);
		} else {
			// Non-empty column:
			$p = $p0 + $ad[$i];
			$tile_column = "";
			while (strlen ($tile_column) < $height_pixels * $bpp)
			{
				$t = getint1 ($f, $p);
				if ($t & 0x80)
				{
					// Append empty rows:
					$tile_column .= str_repeat ($empty_row, $t & 0x7F);
				} else {
					// Copy rows:
					$tile_column .= getstr ($f, $p, $bpp * $t);
				}
			}
		}
		$tiles .= $tile_column;
	}

	/**********************/
	/*** CREATE TILEMAP ***/
	/**********************/
	/*
	 * 0  8 ... 40
	 * 1  9 ... 41
	 * :  :      :
	 * 7 15 ... 47
	 */
	$map = "";
	for ($y = 0; $y < $height_tiles; $y++) // for each tile row
	{
		for ($x = 0; $x < $wh; $x++) // for each tile column
		{
			$map .= int22str($y + $height_tiles * $x);
		}
	}

	/******************/
	/*** CREATE BMP ***/
	/******************/

	$fn_bmp = sprintf ('exp1%s%04d.bmp', $GLOBALS['slash'], $id);
	make_snes_map($map, $tiles, $fn_bmp, $wh, $height_tiles, $bpp, $pal);
}

/*****************************************************************************/
function export_level ($f, &$p, $id)
/*****************************************************************************/
{
	/* Description: Export a level.
	 * Parameters:
	 * - $f: ROM image.
	 * - $p: Offset.
	 * - $id: Resource index.
	 * Returns: Nothing.
	 */

	$level = decode_level ($f, $p);

	$fn = sprintf ('exp0%s%02d.bin', $GLOBALS['slash'], $id);
	$fp_out = fopen ($fn, "wb");
	fwrite ($fp_out, $level);
	fclose ($fp_out);
}

?>